# The "transform" module (this module) is provided mostly for convenience
# and demonstration purposes.  It is hoped that users will create their
# own transform modules that build on top of cvexp rather than relying
# on this module alone to do all transformations.

# WARNING: Unlike the builders module, the data structures in this module
#          might not be purely functional.  This will probably cause grief.

from collections import defaultdict

import vocabulary as vo
import builders

import sympy
_s_zero = 0 * sympy.Symbol('x')
_s_eq = sympy.Function('sympy.==') # SymPy does weird things with ==
_s_x = sympy.Symbol('x')
_s_y = sympy.Symbol('y')
_s_le = (_s_x <= _s_y).func
_s_lt = (_s_x < _s_y).func
del _s_x
del _s_y

def __expr_to_sympy_term_rewrite_expand(expr):
    self = __expr_to_sympy_term_rewrite_expand
    eval_funcs = {
        vo.CONST    : lambda a: _s_zero + a,
        vo.VAR      : lambda a: sympy.Symbol(a),
        # Comparison operations:
        vo.EQ       : lambda a, b: _s_eq(a.expand(), b.expand()),
        vo.NE       : lambda a, b: (a != b).expand(),
        vo.GT       : lambda a, b: (a > b).expand(),
        vo.LT       : lambda a, b: (a < b).expand(),
        vo.GE       : lambda a, b: (a >= b).expand(),
        vo.LE       : lambda a, b: (a <= b).expand(),

        # Math operations:
        vo.ADD      : lambda a, b: a + b,
        vo.SUBTRACT : lambda a, b: a - b,
        vo.MULTIPLY : lambda a, b: a * b,
        vo.DIVIDE   : lambda a, b: a / b,
        #def __truediv__(self, other):   return Expr.function()
        #def __floordiv__(self, other):  return Expr.function()
        #def __mod__(self, other):       return Expr.function()
        #def __divmod__(self, other):    return Expr.function()
        vo.POWER    : lambda a, b: a ** b,

        ## Bit operations:
        #('g.<<',  2): lambda a, b: a << b,
        #('g.>>',  2): lambda a, b: a >> b,
        #('g.&',   2): lambda a, b: a & b,
        #('g.^',   2): lambda a, b: a ^ b,
        #('g.|',   2): lambda a, b: a | b,

        # ...and still more operations:
        vo.NEGATE   : lambda a: -a,
        ## Python calls this whenever abs() is called:
        vo.ABSOLUTE_VALUE: lambda a: sympy.abs(a.expand()),
        #('g.~',   1): lambda a: ~a,

        # XXX: this is starting to get a little out of control:
        vo.SIN      : lambda a: sympy.sin(a),
        vo.COS      : lambda a: sympy.cos(a),
        #('g.&&',  2): lambda a, b: a and b,
        #('g.||',  2): lambda a, b: a or b,
    }   

    if expr.definition.is_terminal:
        return eval_funcs[expr.definition](*expr.args)
    else:
        return eval_funcs[expr.definition](*map(self, expr.args))


def __expr_to_sympy(expr):
    self = __expr_to_sympy
    eval_funcs = {
        vo.CONST    : lambda a: _s_zero + a,
        vo.VAR      : lambda a: sympy.Symbol(a),
        # Comparison operations:
        vo.EQ       : lambda a, b: _s_eq(a, b),
        vo.NE       : lambda a, b: (a != b),
        vo.GT       : lambda a, b: (a > b),
        vo.LT       : lambda a, b: (a < b),
        vo.GE       : lambda a, b: (a >= b),
        vo.LE       : lambda a, b: (a <= b),

        # Math operations:
        vo.ADD      : lambda a, b: a + b,
        vo.SUBTRACT : lambda a, b: a - b,
        vo.MULTIPLY : lambda a, b: a * b,
        vo.DIVIDE   : lambda a, b: a / b,
        #def __truediv__(self, other):   return Expr.function()
        #def __floordiv__(self, other):  return Expr.function()
        #def __mod__(self, other):       return Expr.function()
        #def __divmod__(self, other):    return Expr.function()
        vo.POWER    : lambda a, b: a ** b,

        ## Bit operations:
        #('g.<<',  2): lambda a, b: a << b,
        #('g.>>',  2): lambda a, b: a >> b,
        #('g.&',   2): lambda a, b: a & b,
        #('g.^',   2): lambda a, b: a ^ b,
        #('g.|',   2): lambda a, b: a | b,

        # ...and still more operations:
        vo.NEGATE   : lambda a: -a,
        ## Python calls this whenever abs() is called:
        vo.ABSOLUTE_VALUE: lambda a: sympy.abs(a),
        #('g.~',   1): lambda a: ~a,

        # XXX: this is starting to get a little out of control:
        vo.SIN      : lambda a: sympy.sin(a),
        vo.COS      : lambda a: sympy.cos(a),
        #('g.&&',  2): lambda a, b: a and b,
        #('g.||',  2): lambda a, b: a or b,
    }   

    if expr.definition.is_terminal:
        return eval_funcs[expr.definition](*expr.args)
    else:
        return eval_funcs[expr.definition](*map(self, expr.args))


def __sympy_to_expr(e):
    self = __sympy_to_expr
    #
    if e.is_Add:        return sum(map(self, e.args))
    if e.is_Mul:        return reduce(lambda a,b: a*b, map(self, e.args))
    if e.is_Pow:        return self(e.args[0]) ** self(e.args[1])
    if e.func == _s_le: return self(e.args[0]) <= self(e.args[1])
    if e.func == _s_lt: return self(e.args[0]) <  self(e.args[1])
    if e.func == _s_eq: return self(e.args[0]) == self(e.args[1])
    if e.is_Symbol:     return builders.var(e.name)
    # else: assume constant
    if e.is_Real:       return builders.const(float(e))
    if e.is_Integer:    return builders.const(int(e))
    raise Exception('ERROR: %s: %s' % (type(e), e))
    assert(False)

def expand_polynomial(expr):
    s = __expr_to_sympy(expr).expand()
    if not s.is_polynomial():
        print expr_to_scheme_(expr)
        print 'ERROR:', s
        assert(False)
    e = __sympy_to_expr(s)
    return e

def expr_to_scheme_(expr):
    if expr.definition.is_terminal:
        return expr.args[0]
    else:
        args = map(expr_to_scheme_, expr.args)
        if   expr.definition is vo.ADD:
            return '(+ %s %s)' % (args[0], args[1])
        elif expr.definition is vo.SUBTRACT:
            return '(- %s %s)' % (args[0], args[1])
        elif expr.definition is vo.MULTIPLY:
            return '(* %s %s)' % (args[0], args[1])
        elif expr.definition is vo.DIVIDE:
            return '(/ %s %s)' % (args[0], args[1])
        elif expr.definition is vo.POWER:
            # XXX: this isn't the right operator...
            return '(** %s %s)' % (args[0], args[1])
        else:
            assert(False)

class TermNormalForm(object):

    def __init__(self, term_expr):
        self.__expr__ = term_expr
        term = term_expr

        # Extract multiplicands
        # {
        stack = [term]
        multiplicands = []
        while stack:
            top = stack.pop()
            if top.definition is vo.MULTIPLY:
                stack.append(top.args[0])
                stack.append(top.args[1])
            else:
                multiplicands.append(top)
        del stack
        # }

        # ...there should probably be a comment on this line...
        # {
        coefficient = 1
        exponents = defaultdict(lambda: 0)
        for m in multiplicands:
            if m.definition is vo.VAR:
                exponents[m] += 1
            elif m.definition is vo.POWER and \
                 m.args[0].definition is vo.VAR and \
                 m.args[1].definition is vo.CONST:
                var = m.args[0]
                const = m.args[1]
                if const != 0: exponents[var] += const.args[0]
            elif m.definition is vo.CONST:
                coefficient *= m.args[0]
        # }

        # {
        self.__coefficient__ = coefficient
        self.__degree__ = sum(exponents.values())
        self.__exponents__ = exponents
        # }

    def __eq__(self, other):
        return self.coefficient == other.coefficient and \
               self.exponents == other.exponents

    @property
    def coefficient(self):
        return self.__coefficient__

    @property
    def degree(self):
        return self.__degree__

    @property
    def exponents(self):
        return self.__exponents__

    @property
    def to_expr(self):
        raise Exception('Not Implemented')

class PolynomialNormalForm(object):

    def __init__(self, expr):
        expr = expand_polynomial(expr)

        # Extract addends
        # {
        stack = [expr]
        addends = []
        while stack:
            top = stack.pop()
            if top.definition is vo.ADD:
                stack.append(top.args[0])
                stack.append(top.args[1])
            else:
                addends.append(top)
        del stack
        # }

        terms = map(TermNormalForm, addends)
        terms = [x for x in terms if x.coefficient != 0]

        degree_term = defaultdict(lambda: set())
        for term in terms:
            degree_term[term.degree].add(term)
        # TODO: remap set to frozenset

        self.__terms__ = terms
        self.__degree_term__ = degree_term
        self.__degree__ = max(degree_term.keys())

    @property
    def terms(self):
        return self.__terms__

    @property
    def degree(self):
        return self.__degree__

    def terms_by_degree(self, degree):
        if degree in self.__degree_term__: return self.__degree_term__[degree]
        else:                              return frozenset([])

    @property
    def to_expr(self):
        raise Exception('Not Implemented')


def cull_vars(expr):
    result = []
    stack = [expr]
    while stack:
        top = stack.pop()
        if not top.definition.is_terminal:
            for arg in top.args: stack.append(arg)
        elif top.definition is vo.VAR:
            result.append(top)
    return result

class OptimizationNormalForm(object):

    def __init__(self, *exprs):

        self.le        = []
        self.eq        = []
        self.minimize  = []
        self.vars      = set()
        self.int_vars  = set()

        #self.lt        = []
        #self.ne        = []

        for e in exprs:
            for v in cull_vars(e): self.vars.add(v)
            if False: pass
            elif e.definition is vo.LE:
                self.le.append(PolynomialNormalForm(e.args[0] - e.args[1]))
            elif e.definition is vo.GE:
                self.le.append(PolynomialNormalForm(e.args[1] - e.args[0]))
            #elif e.definition is vo.LT:
            #elif e.definition is vo.GT:
            elif e.definition is vo.EQ:
                self.eq.append(PolynomialNormalForm(e.args[0] - e.args[1]))
            #elif e.definition is vo.NE
            elif e.definition is vo.MINIMIZE:
                self.minimize.append(PolynomialNormalForm(e.args[0]))
            elif e.definition is vo.MAXIMIZE:
                self.minimize.append(PolynomialNormalForm(-e.args[0]))
            elif e.definition is vo.INTEGER:
                self.int_vars.add(e.args[0])
            else:
                raise Exception('Invalid expression for optimization: %s: %s' %
                                (e.definition,
                                 repr(e)
                                )
                               )

        #self.real_vars = set(vars) - set(int_vars)

def test():
    import time
    t0 = time.time()

    zero = builders.const(0)
    x = builders.var('x')
    y = builders.var('y')

    ex = zero + 1
    poly = PolynomialNormalForm(ex)
    assert(poly.degree == 0)

    ex = zero + x
    poly = PolynomialNormalForm(ex)
    assert(poly.degree == 1)

    a = x
    b = 3*y
    c = 2*x**2*y**3*3
    aa = TermNormalForm(a)
    bb = TermNormalForm(b)
    cc = TermNormalForm(c)
    #
    assert(aa.coefficient == 1)
    assert(aa.exponents[x] == 1)
    #
    assert(bb.coefficient == 3)
    assert(bb.exponents[y] == 1)
    #
    assert(cc.coefficient == 6)
    assert(cc.exponents[x] == 2)
    assert(cc.exponents[y] == 3)
    #
    ex = a + b + c
    poly = PolynomialNormalForm(ex)
    #
    assert(poly.degree == 5)
    assert(poly.terms_by_degree(0) == set())
    # XXX: this is awkward:
    assert(len(poly.terms_by_degree(1)) == 2)
    assert(max([aa == x for x in poly.terms_by_degree(1)]))
    assert(max([bb == x for x in poly.terms_by_degree(1)]))
    assert(poly.terms_by_degree(2) == set())
    assert(poly.terms_by_degree(3) == set())
    assert(poly.terms_by_degree(4) == set())
    assert(len(poly.terms_by_degree(5)) == 1)
    assert(max([cc == x for x in poly.terms_by_degree(5)]))
    assert(poly.terms_by_degree(6) == set())

    t1 = time.time()
    print 'Test completed in [seconds]:', t1 - t0

if __name__ == '__main__':
    test()
