# Might need to do a trick like this to get it working in Ubuntu:
# LD_PRELOAD='/usr/lib/lp_solve/liblpsolve55.so' python

from builders import generate_uniqid
import _translate
import lpsolve55
from lpsolve55 import lpsolve
from lpsolve55 import IMPORTANT
from lpsolve55 import OPTIMAL
from lpsolve55 import SUBOPTIMAL
from lpsolve55 import LE
from lpsolve55 import EQ

def solve(exprs, out_filename=None):
    prob = _translate.OptimizationNormalForm(*exprs)

    lp = lpsolve('make_lp', 0, len(prob.vars))
    lpsolve('set_verbose', lp, IMPORTANT)

    var_i = dict([(v, i) for i, v in enumerate(prob.vars)])
    i_var = dict([(i, v) for i, v in enumerate(prob.vars)])

    # Minimize:     q^T x
    # Subject to:   G x <= h
    #               A x = b

    # XXX: DRY (with cvxopt):
    def extract_coefficients(poly):
        assert(poly.degree == 1)
        constant = 0
        coeffs = [0] * len(var_i)
        for term in poly.terms:
            if term.degree == 0:
                constant = term.coefficient
            else:
                [(var, exp)] = term.exponents.items()
                col = var_i[var]
                coeffs[col] = term.coefficient
                # Make sure this is really a first order polynomial:
                assert(len(term.exponents) == 1)
                assert(exp == 1)
        return constant, coeffs

    # G x <= h constraints:
    for ci, constraint in enumerate(prob.le):
        constant, coeffs = extract_coefficients(constraint)
        lpsolve('add_constraint', lp, coeffs, LE, -constant)

    # A x = b constraints:
    for ci, constraint in enumerate(prob.eq):
        constant, coeffs = extract_coefficients(constraint)
        lpsolve('add_constraint', lp, coeffs, EQ, -constant)

    # integer constraints:
    for v in prob.int_vars:
        col = var_i[v] + 1 # column numbers start at 1
        lpsolve('set_int', lp, col, 1)

    # q^T x objective function:
    assert(len(prob.minimize) == 1) # don't know how to satisfy more than 1
    constant, min_coeffs = extract_coefficients(prob.minimize[0])
    #max_coeffs = [-x for x in min_coeffs]
    lpsolve('set_obj_fn', lp, min_coeffs)

    #lpsolve('write_lp', lp, generate_uniqid('debug', '.lp'))
    if out_filename:
        print out_filename
        lpsolve('write_lp', lp, out_filename)

    ret = lpsolve('solve', lp)

    if ret == OPTIMAL: # what about SUBOPTIMAL?
        sol = lpsolve('get_variables', lp)[0]
        result = { }
        for v, i in var_i.iteritems():
            if v in prob.int_vars:
                result[v] = int(sol[i])
            else:
                result[v] = sol[i]
        return result
    else:
        return None

def test():
    from builders import var
    from builders import minimize
    from builders import integer

    # TODO: make sure that the P matrix is right (especially the diagonals).
    #       do this by trying to balance out minimizations of x^2 + x*y
    # TODO: try out maximize
    # TODO: try out integer constraints

    def fuzzy_equal(a, b, fuzz):
        return (a - fuzz) < b < (a + fuzz)

    X = var('X')
    Y = var('Y')

    if True:
        sol = solve((
                    Y == X,
                    Y >= 10 - X,
                    minimize(Y),
                   ), out_filename='lpsolve001.lp')
        print 'X =', sol[X]
        print 'Y =', sol[Y]
        assert(fuzzy_equal(sol[X], 5, 0.000001))
        assert(fuzzy_equal(sol[Y], 5, 0.000001))

    if True:
        sol = solve((
                    Y == X,
                    Y >= 9.8 - X,
                    minimize(Y),
                   ), out_filename='lpsolve002.lp')
        print 'X =', sol[X]
        print 'Y =', sol[Y]
        assert(fuzzy_equal(sol[X], 4.9, 0.000001))
        assert(fuzzy_equal(sol[Y], 4.9, 0.000001))

    if True:
        sol = solve((
                    Y + 0.1 == X,
                    Y >= 9.8 - X,
                    integer(Y),
                    minimize(Y),
                   ), out_filename='lpsolve003.lp')
        print 'X =', sol[X]
        print 'Y =', sol[Y]
        assert(fuzzy_equal(sol[X], 5.1, 0.000001))
        assert(fuzzy_equal(sol[Y], 5.0, 0.000001))

if __name__ == '__main__':
    test()
