basic         = 'http://aclevername.com/ns/20090213/cvexp/basic/#'
constraint    = 'http://aclevername.com/ns/20090213/cvexp/constraint/#'
optimization  = 'http://aclevername.com/ns/20090213/cvexp/optimization/#'
logic         = 'http://aclevername.com/ns/20090213/cvexp/logic/#'
nice_function = 'http://aclevername.com/ns/20090213/cvexp/well-known-function/#'

class AnyArity(object): pass
class IgnoreArity(object): pass

class Definition(object):


    flyweight = { }

    def __init__(self, word, arity, terminal_argument, magic):
        assert(magic == 42) # to discourage people from calling __init__
        self.word  = word
        self.arity = arity
        self.is_terminal = terminal_argument

    @classmethod
    def define(cls, word, arity=IgnoreArity, terminal_argument=False):
        word = intern(word)
        key = word

        if key in cls.flyweight:
            x = cls.flyweight[key]
            if arity is not IgnoreArity:
                assert(x.arity == arity)
        else:
            assert(arity is not IgnoreArity)
            x = cls(word, arity, terminal_argument, 42)
            cls.flyweight[key] = x

        return x

    def __repr__(self):
        return 'Definition(word=%s, arity=%s, is_terminal=%s)' % (repr(self.word), str(self.arity), str(self.is_terminal))

def define(word, arity=IgnoreArity, terminal_argument=False):
    return Definition.define(word, arity, terminal_argument)

EQ  = define(basic + 'equal-to',       2) # ==
NE  = define(basic + 'not-equal',      2) # !=
GT  = define(basic + 'greater-than',   2) # >
LT  = define(basic + 'less-than',      2) # <
GE  = define(basic + 'greater-equal',  2) # >=
LE  = define(basic + 'less-equal',     2) # <=

ADD      = define(basic + 'add',      2) # +
SUBTRACT = define(basic + 'subtract', 2) # -
MULTIPLY = define(basic + 'multiply', 2) # *
DIVIDE   = define(basic + 'divide',   2) # /

POWER    = define(basic + 'power',    2) # **

NEGATE   = define(basic + 'negate',   1) # -x

LEFT_SHIFT  = define(basic + 'left-shift',  2) # <<
RIGHT_SHIFT = define(basic + 'right-shift', 2) # >>
BINARY_AND  = define(basic + 'binary-and',  2) # &
BINARY_OR   = define(basic + 'binary-or',   2) # |
BINARY_XOR  = define(basic + 'binary-xor',  2) # ^
INVERT      = define(basic + 'invert',      1) # ~x

LOGICAL_AND = define(basic + 'logical-and', 2) # &&
LOGICAL_OR  = define(basic + 'logical-or',  2) # ||

# TODO: assignment operator?

ABSOLUTE_VALUE = define(nice_function + 'abs', 1)
SIN            = define(nice_function + 'sin', 1)
COS            = define(nice_function + 'cos', 1)

CONST = define(basic + 'const', 1, terminal_argument=True)
VAR   = define(basic + 'var',   1, terminal_argument=True)
# Deliberately ambiguous.  Can be either:
#
# - A logic variable (whose value must be consistent/identical everywhere
#   one bound)
# - A state variable (whose value can keep changing)

INTEGER = define(constraint + 'integer', 1)

# For optimization problems (e.g. linear programming):
MINIMIZE = define(optimization + 'minimize', 1)
MAXIMIZE = define(optimization + 'maximize', 1)
